from typing import Any

from pydantic import Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from snakestack.logging.enumerators import LogFilters, LogFormatter


class LoggingSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="SNAKESTACK_LOGGING_")

    level: str = Field(
        default="INFO",
        description="Logging level for the application (e.g., DEBUG, INFO, WARNING, ERROR)."
    )

    name: str = Field(
        default="__main__",
        description="Name of the application (e.g., snakestack)."
    )

    lib_level: str = Field(
        default="WARNING",
        description="Logging level for the third libraries (e.g., INFO, WARNING, ERROR)."
    )

    formatter: str = Field(
        default="default",
        description="Default formatter to be used in the logging configuration (e.g., default, json, with_request_id)."
    )

    filters: list[str] | str  = Field(
        default="request_id",
        description="Comma-separated list of default filters to be applied to log records (e.g., request_id, excluded_name, filter_paths)."
    )

    filter_excluded_name: list[str] | None = Field(
        default=None,
        description="Logger name or pattern to exclude from logging output (e.g., 'exclude.me' to suppress logs from that logger)."
    )

    filter_paths: list[str] | None = Field(
        default=None,
        description="Comma-separated list of paths to filter log records (e.g., '/healthcheck' to suppress logs)."
    )

    @field_validator("formatter")
    @classmethod
    def validate_log_formatter(cls, v: str) -> str:
        allowed = [str(member.value) for member in LogFormatter]
        if v not in allowed:
            raise ValueError(f"Invalid formatter '{v}'. Must be one of: {', '.join(allowed)}.")
        return v

    @field_validator("filters")
    @classmethod
    def validate_log_filter(cls, v: str) -> list[str]:
        allowed = [str(member.value) for member in LogFilters]

        if isinstance(v, list):
            if not all([item in allowed for item in v]):
                raise ValueError(f"Invalid filter '{v}'. Must be one of: {', '.join(allowed)}.")
            return v

        if not all ([item.strip() in allowed for item in v.split(",")]):
            raise ValueError(f"Invalid filter '{v}'. Must be one of {', '.join(allowed)}.")

        return v.split(",")

    @field_validator("filter_excluded_name", mode="before")
    @classmethod
    def convert_filter_excluded_name(cls, v: Any) -> list[str]:
        if not isinstance(v, str):
            return []

        parts = [p.strip() for p in v.split(",") if v.split()]
        return [p for p in parts if p]

    @field_validator("filter_paths", mode="before")
    @classmethod
    def convert_filter_paths(cls, v: Any) -> list[str]:
        if not isinstance(v, str):
            return []

        parts = [p.strip() for p in v.split(",") if v.split()]
        return [p for p in parts if p]

    @field_validator("level", "lib_level")
    @classmethod
    def validate_log_level(cls, v: str) -> str:
        allowed = {"CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"}
        value = v.upper().strip()
        if value not in allowed:
            raise ValueError(f"Invalid log level '{v}'. Must be one of {', '.join(allowed)}.")
        return value



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class OtelSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="SNAKESTACK_OTEL_")

    disabled: bool = Field(
        default=False,
        description=(
            "Disables only the OpenTelemetry instrumentation provided by the SnakeStack library. "
            "This allows external OpenTelemetry configurations (e.g., opentelemetry-bootstrap or custom distro) "
            "to take full control over instrumentation without duplication. "
            "Equivalent to setting the environment variable SNAKESTACK_OTEL_DISABLED=true."
        )
    )



from enum import Enum, StrEnum


class LogLevel(int, Enum):
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOT_SET = 0


class LogFormatter(StrEnum):
    DEFAULT = "default"
    JSON = "json"
    WITH_REQUEST_ID = "with_request_id"


class LogFilters(StrEnum):
    REQUEST_ID = "request_id"
    EXCLUDED_NAME = "excluded_name"
    PATH_FILTER = "path_filter"



import importlib
from typing import Any


def load_module(name: str) -> Any:
    if "." not in name:
        raise AttributeError(f"Invalid module [{name}]")

    module_name, identifier = name.rsplit(".", 1)
    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError as error:
        raise AttributeError(f"Module not found {error}") from error
    return getattr(module, identifier)



from typing import Any, Self

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class MongoDBSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="SNAKESTACK_MONGODB_")

    uri: str = Field(
        default="mongodb://localhost:27017",
        description=""
    )
    dbname: str = Field(
        default="snakestack",
        description=""
    )
    user: str | None = Field(
        default=None,
        description=""
    )
    password: str | None = Field(
        default=None,
        description=""
    )
    db_auth_source: str = Field(
        default="admin",
        description=""
    )

    def get_host(self: Self) -> str:
        return f"{self.uri}/{self.dbname}"

    def get_kwargs(self: Self) -> dict[str, Any]:
        config = {}
        if self.user and self.password:
            config["username"] = self.user
            config["password"] = self.password
            config["authSource"] = self.db_auth_source

        return config



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class CacheSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="SNAKESTACK_CACHE_")
    host: str = Field(
        default="localhost",
        description=""
    )
    port: int = Field(
        default=6379,
        description=""
    )
    db: int = Field(
        default=0,
        description=""
    )
    password: str | None = Field(
        default=None,
        description=""
    )



import logging
import logging.config
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from snakestack.config import SnakeStackSettings


def setup_logging(
    settings: "SnakeStackSettings"
) -> None:
    filters = {
        "request_id": {"()": "snakestack.logging.filters.RequestIdFilter"},
        "excluded_name": {
            "()": "snakestack.logging.filters.ExcludeLoggerFilter",
            "excluded_name": settings.logging.filter_excluded_name
        },
        "path_filter": {
            "()": "snakestack.logging.filters.PathFilter",
            "paths": settings.logging.filter_paths
        }
    }

    third_party_formatter = (
        "default"
        if settings.logging.formatter == "with_request_id"
        else settings.logging.formatter
    )

    handlers = {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": settings.logging.formatter,
            "filters": settings.logging.filters,
        },
        "third_party_console": {
            "class": "logging.StreamHandler",
            "formatter": third_party_formatter
        },
    }

    formatters = {
        "default": {
            "format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"
        },
        "with_request_id": {
            "format": (
                "%(asctime)s [%(levelname)s] [req_id=%(request_id)s] "
                "%(name)s: %(message)s"
            )
        },
        "json": {
            "()": "snakestack.logging.formatters.JsonFormatter"
        }
    }

    logging_config: dict[str, Any] = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": formatters,
        "handlers": handlers,
        "filters": filters,
        "loggers": {
            settings.logging.name: {
                "level": settings.logging.level,
                "handlers": ["console"],
                "propagate": False,
            },
            "snakestack": {
                "level": settings.logging.level,
                "handlers": ["console"],
                "propagate": False,
            }
        },
        "root": {
            "level": settings.logging.lib_level,
            "handlers": ["console"]
        }
    }

    logging.config.dictConfig(logging_config)



import asyncio
import logging
import time
from typing import Awaitable, Callable, Self, TypeAlias

try:
    from google.cloud.pubsub_v1 import SubscriberClient
    from google.cloud.pubsub_v1.subscriber.message import Message
except ImportError:
    raise RuntimeError("Pubsub extra is not installed. Run `pip install snakestack[pubsub]`.")

from snakestack.model.lenient import LenientModel
from snakestack.pubsub.subscriber.types import AsyncQueueMessagePubSub, DataPubSubType

logger = logging.getLogger(__name__)


class BatchSettingsModel(LenientModel):
    size: int
    timeout: float
    name: str
    subscriber_client: SubscriberClient | None = None
    subscription_path: str | None = None


class BatchingWorker:
    def __init__(
        self: Self,
        *,
        callback: Callable[[list[tuple[Message, DataPubSubType]], BatchSettingsModel], Awaitable[None]],
        batch_settings: BatchSettingsModel
    ) -> None:
        self.buffer: list[tuple[Message, DataPubSubType]] = []
        self.last_flush = time.monotonic()
        self.callback = callback
        self.batch_settings = batch_settings
        self._queue: AsyncQueueMessagePubSub | None = None

    async def run(
        self: Self,
        queue: AsyncQueueMessagePubSub
    ) -> None:
        self._queue = queue
        try:
            logger.debug(f"Starting batch [{self.batch_settings.name}] "
                         f"- Timeout [{self.batch_settings.timeout}] "
                         f"- Size [{self.batch_settings.size}]")
            while True:
                try:
                    timeout = self._get_timeout()
                    item = await asyncio.wait_for(queue.get(), timeout=timeout)
                    self.buffer.append(item)
                except asyncio.TimeoutError:
                    pass

                conditional_size = self._get_size_conditional()
                conditional_time = self._get_time_conditional()
                if any([conditional_time, conditional_size]):
                    await self._flush()
        except asyncio.CancelledError:
            await self._flush()
            raise

    async def _flush(self: Self) -> None:
        if not self.buffer:
            return

        logger.info(f"[{self.batch_settings.name}] Executing batch with {len(self.buffer)} messages")
        try:
            await self.callback(self.buffer, self.batch_settings)
        except Exception:
            logger.exception("Failed to process batch")
        finally:
            if self._queue:
                for _ in range(len(self.buffer)):
                    self._queue.task_done()
            self.buffer.clear()
            self.last_flush = time.monotonic()

    def _get_size_conditional(self: Self) -> bool:
        return len(self.buffer) >= self.batch_settings.size

    def _get_time_conditional(self: Self) -> bool:
        return (time.monotonic() - self.last_flush) >= self.batch_settings.timeout

    def _get_timeout(self: Self) -> float:
        return max(
            self.batch_settings.timeout - (time.monotonic() - self.last_flush), 0.1
        )

BatchesType: TypeAlias = dict[str, BatchingWorker]



from datetime import UTC, datetime
from logging import Formatter, LogRecord
from typing import Any, Self

import orjson

from snakestack.logging.encoders import safe_jsonable_encoder


class JsonFormatter(Formatter):
    def format(
        self: Self,
        record: LogRecord,
        *args: Any,
        **kwargs: Any
    ) -> str:
        Formatter.format(self, record)
        return orjson.dumps(
            self._create_log(record),
            default=safe_jsonable_encoder
        ).decode("utf-8")

    def _create_log(self: Self, record: LogRecord) -> dict[str, Any]:
        message: dict[str, Any] = {
            "time": datetime.fromtimestamp(record.created, tz=UTC).isoformat(),
            "level": record.levelname,
            "pid": record.process,
            "name": f"{record.name}:{record.lineno}",
            "message": record.getMessage()
        }

        if record.exc_info:
            message["exc_info"] = self.formatException(record.exc_info)

        if hasattr(record, "request_id") and record.request_id:
            message.update({
                "request": {
                    "id": record.request_id
                }
            })

        if hasattr(record, "otelSpanID") and hasattr(record, "otelTraceID"):
            message.update({
                "open_telemetry": {
                    "trace_id": record.otelTraceID,
                    "span_id": record.otelSpanID
                }
            })

        if hasattr(record, "extra_data") and record.extra_data:
            message.update({
                "extra": record.extra_data
            })

        return message



import logging
from typing import Any

try:
    from pymongo import AsyncMongoClient
    from pymongo.asynchronous.database import AsyncDatabase
    from pymongo.errors import PyMongoError
except ImportError:
    raise RuntimeError("Mongodb extra is not installed. Run `pip install snakestack[mongodb]`.")

from snakestack.config import SnakeStackSettings


class DB:
    client: AsyncMongoClient[Any]

db = DB()

logger = logging.getLogger(__name__)


async def open_mongodb_connection(
    settings: SnakeStackSettings
) -> None:
    logger.debug("Connecting to mongodb...")
    if not settings.mongo.uri.startswith("mongodb"):
        raise ValueError("Invalid MongoDB URL configured.")
    db.client = AsyncMongoClient(
        settings.mongo.get_host(),
        **settings.mongo.get_kwargs()
    )
    logger.debug("Connecting to mongodb successful.")


async def close_mongodb_connection() -> None:
    logger.debug("Closing connection with mongodb...")
    if db.client:
        await db.client.close()
    logger.debug("Connection with mongodb is closed.")


def get_database(
    settings: SnakeStackSettings
) -> AsyncDatabase[Any]:
    if not db.client:
        raise RuntimeError("Connection with mongodb is not starting.")
    return db.client[settings.mongo.dbname]


async def ping() -> bool:
    if not db.client:
        raise RuntimeError("Connection with mongodb is not starting.")
    try:
        await db.client.admin.command("ping")
        return True
    except PyMongoError:
        return False



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class PubSubSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="SNAKESTACK_PUBSUB_")
    project_id: str = Field(
        default="snakestack-project",
        description=""
    )
    subscription_name: str | None = Field(
        default=None,
        description=""
    )
    topic_name: str | None = Field(
        default=None,
        description=""
    )



import logging
from typing import Self

from snakestack.logging.contexts import get_request_id


class RequestIdFilter(logging.Filter):
    def filter(self: Self, record: logging.LogRecord) -> bool:
        record.request_id = get_request_id()
        return True

class ExcludeLoggerFilter(logging.Filter):
    def __init__(self: Self, excluded_name: list[str]) -> None:
        super().__init__()
        cleaned = [n.strip() for n in excluded_name if n and n.strip()]
        self.excluded = tuple(sorted(set(cleaned)))

    def filter(self: Self, record: logging.LogRecord) -> bool:
        name = record.name
        return not any(name == n or name.startswith(f"{n}." ) for n in self.excluded)


class PathFilter(logging.Filter):
    def __init__(self: Self, paths: list[str] | None) -> None:
        super().__init__()
        self.paths = paths or []

    def filter(self: Self, record: logging.LogRecord) -> bool:
        if hasattr(record, "path_url"):
            url = record.path_url
            return not any(url.startswith(path) for path in self.paths)

        return True



from functools import lru_cache

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from snakestack import version
from snakestack.cache.settings import CacheSettings
from snakestack.logging.settings import LoggingSettings
from snakestack.mongodb.settings import MongoDBSettings
from snakestack.pubsub.settings import PubSubSettings
from snakestack.telemetry.settings import OtelSettings


class SnakeStackSettings(BaseSettings):
    model_config = SettingsConfigDict(
        env_file=".env",
        env_prefix="SNAKESTACK_APP_",
        env_file_encoding="utf-8",
        env_nested_delimiter="__",
    )

    version: str = Field(
        default=version.__version__,
        description=""
    )

    cache: CacheSettings = Field(default_factory=CacheSettings)
    logging: LoggingSettings = Field(default_factory=LoggingSettings)
    mongo: MongoDBSettings = Field(default_factory=MongoDBSettings)
    pubsub: PubSubSettings = Field(default_factory=PubSubSettings)
    otel: OtelSettings = Field(default_factory=OtelSettings)


@lru_cache
def get_settings() -> SnakeStackSettings:
    return SnakeStackSettings()


settings = get_settings()
